---
driver number: 0x00009
---

# Read Only State

## Overview

This capsule provides read only state to userspace applications.
This is similar to the Linux vDSO syscalls.

The benefit of using these is that applications can avoid the context
switch overhead of traditional syscalls by just reading the value from
memory.

The value will only be as accurate as the last time the application was
switched to by the kernel.

The layout of the read only state in the allow region depends on the
version. Userspace can use `command 0` to get the version information.

Versions are backwards compatible, that is new versions will only add
fields, not remove existing ones or change the order.

Version 1:
  |-------------------------|
  |    Switch Count (u32)   |
  |-------------------------|
  |   Pending Tasks (u32)   |
  |-------------------------|
  |                         |
  |     Time Ticks (u64)    |
  |-------------------------|

`Switch Count`: The number of context switches that have occured, per app.
`Pending Tasks`: The number of currently pending tasks scheduled for this
app. This is the number of upcalls that will be called when the app yields.
`Time Ticks`: The current number of ticks that have occured.

Example C code to safely read the 64-bit timer value is included below,
where `ptr` is a `uint32_t*` pointing to the ROS memory region.

```c
  volatile uint32_t* ptr; // Points to ROS memory region
  uint32_t time_ticks_high, time_ticks_low;
  uint32_t switch_count = 0;

  do {
    // Start our attempt at an atomic read
    switch_count = ptr[0];
    // Set the high bytes the value in memory
    time_ticks_high = ptr[3];
    // Read the low bytes
    time_ticks_low = ptr[2];
    // Check if there was a context switch while reading this value out, if so, retry
  } while (switch_count != ptr[0]);
```

## Data Consistency

Before returning from the kernel to userspace all ROS fields will be updated.
This ensures that at all times when an application is running, the ROS fields
are valid and consistent.

If userspace wants to read multiple values and ensure consistency between reads
the `count` field should be used to ensure that the kernel didn't run between
reads.

The `count` field is updated on ever context switch when the app has provided a
valid userspace readable buffer.

## Allow Shared

  * Description allow_shared() is used to setup the read only buffer:

  * ### Allow Read/Write Number: 0

    **Description**: Memory Buffer.

    **Argument 1**: Slice into which the read only syscall values should be stored.

    **Returns**: SUCCESS

## Command

  * Description: command() is used to query the version:

  * ### Command Number: 0

    **Description**: Get Version.

    **Argument 1**: Unused

    **Argument 2**: Unused

    **Returns**: The version of ROS supported by the current kernel.
